require 'torch'   -- torch
require 'image'   -- for image transforms
require 'nn'      -- provides a normalization operator
require 'optim'   -- an optimization package, for online and batch methods

function isnan(x) return x ~= x end

function lapnoise(l, n) 
   local noise = - torch.rand(n):log():mul(l)
   local sign = torch.rand(n):apply(function(x) if x < 0.5 then return 1 else return -1 end end)
   return noise:cmul(sign)
end

function cut(value, range)
   local cvalue = value
   if range > 0 then
      if cvalue < -range then
         cvalue = -range
      elseif cvalue > range then
         cvalue = range
      end
   end
   return cvalue
end

-- options
cmd = torch.CmdLine()

cmd:option('-dataset',           'mnist', 'svhn | mnist')
cmd:option('-dataSizeFrac',      1,       'the fraction of dataset to be used for training')
cmd:option('-model',             'cvn',   'convnet | cvn | linear | mlp | deep')
cmd:option('-method',            'seq',   'seq | fla | asy')
cmd:option('-learningRate',      1e-3,    '')
cmd:option('-learningRateDecay', 1e-7,    '')
cmd:option('-batchSize',         1,       '')
cmd:option('-weightDecay',       0,       '')
cmd:option('-momentum',          0,       '')
cmd:option('-threads',           2,       '')
cmd:option('-netSize',           30,      '')
cmd:option('-shardSizeFrac',     0.01,    'fraction of the training set in each shard')
cmd:option('-uploadFraction',    0.1,     'fraction of parameters to be uploaded after training')
cmd:option('-downloadFraction',  0.1,     'fraction of parameters to be downloaded before training')
cmd:option('-epochFraction',     1,       '')
cmd:option('-epsilon',           0,       'epsilon for dp. 0: disable it')
cmd:option('-delta',             0,       'delta for dp.')
cmd:option('-range',             0,       'cut the gradiants between -range and range. 0: disable it')
cmd:option('-threshold',         0,       'release those whose abs value is greater than threshold. 0: disable it')
cmd:option('-nepochs',           60,     '')
cmd:option('-taskID',            '0',     'the ID associated to the task')
cmd:option('-folder',            'save',  '')

opt = cmd:parse(arg or {})
 
epochfile = paths.concat(opt.folder, 'epoch.' .. opt.taskID)
resultsfile = paths.concat(opt.folder, 'results.' .. opt.taskID)
pfile = paths.concat(opt.folder, 'p.' .. opt.taskID)
paramfile = paths.concat(opt.folder, 'parameters.' .. opt.taskID)
gradstatfile = paths.concat(opt.folder, 'gradstat.' .. opt.taskID)
gradstatpfile = paths.concat(opt.folder, 'gradstatp.' .. opt.taskID)
nupdatesfile = paths.concat(opt.folder, 'nupdates.' .. opt.taskID)
 
print(opt)

-- config torch
torch.setnumthreads(opt.threads)
torch.setdefaulttensortype('torch.FloatTensor')

-- loading dataset
trainData = torch.load('train_' .. opt.dataset)
testData  = torch.load('test_' .. opt.dataset)   

trainSize = math.ceil(opt.dataSizeFrac * trainData.labels:size(1))
testSize  = testData.labels:size(1)
shardSize = math.ceil(opt.shardSizeFrac * trainSize)

trainData.size = function() return trainSize end
testData.size = function() return testSize end
trainData.shardsize = function() return shardSize end

trainData.shard = {}

for ndid = 1, opt.netSize do
   local shffl = torch.randperm(trainData:size())
   trainData.shard[ndid] = shffl[{ {1,trainData:shardsize()} }]
end

print(trainData)
print(testData)

-- logger:add{['%'] = 'trainSize ' .. trainSize}
-- logger:add{['%'] = 'testSize ' .. testSize}
-- logger:add{['%'] = 'shardSize ' .. shardSize}

if opt.dataset == 'svhn' then
   -- model parameters
   nfeats = 3
   width  = 32
   height = 32
   classes = {'1','2','3','4','5','6','7','8','9','0'}

elseif opt.dataset == 'mnist' then
   nfeats = 1
   width  = 32
   height = 32
   classes = {'1','2','3','4','5','6','7','8','9','0'}

   if opt.model == 'deep' then
      trainData.data = trainData.data:squeeze()
      testData.data = testData.data:squeeze()
   end
end

ninputs   = nfeats * width * height
nhiddens  = 128 -- ninputs / 6
nhiddens2 = 64 -- ninputs / 12
noutputs  = 10

nstates = {64,64,128}
filtsize = 5
poolsize = 2
normkernel = image.gaussian1D(7)

-- constructing the model
model = nn.Sequential()

---if paths.filep(modelfile) then
---   model = torch.load(modelfile)
---   print('load model from file')
---else
   if opt.model == 'linear' then
      -- Simple linear model
      model:add(nn.Reshape(ninputs))
      model:add(nn.Linear(ninputs,noutputs))

   elseif opt.model == 'mlp' then
      -- Simple 2-layer neural network, with tanh hidden units
      model:add(nn.Reshape(ninputs))
      model:add(nn.Linear(ninputs,nhiddens))
      model:add(nn.Tanh())
      model:add(nn.Linear(nhiddens,noutputs))

   elseif opt.model == 'deep' then
      -- Deep neural network, with ReLU hidden units
      model:add(nn.Reshape(ninputs))
      model:add(nn.Linear(ninputs,nhiddens))
      model:add(nn.ReLU())
      model:add(nn.Linear(nhiddens,nhiddens2))
      model:add(nn.ReLU())
      model:add(nn.Linear(nhiddens2,noutputs))

      model:add(nn.Linear(nstates[3], noutputs))

   elseif opt.model == 'cvn' then
      if opt.dataset == 'mnist' then
         -- stage 1 : mean suppresion -> filter bank -> squashing -> max pooling
         model:add(nn.SpatialConvolutionMM(1, 32, 5, 5))
         model:add(nn.Tanh())
         model:add(nn.SpatialMaxPooling(3, 3, 3, 3))

         -- stage 2 : mean suppresion -> filter bank -> squashing -> max pooling
         model:add(nn.SpatialConvolutionMM(32, 64, 5, 5))
         model:add(nn.Tanh())
         model:add(nn.SpatialMaxPooling(2, 2, 2, 2))

         -- stage 3 : standard 2-layer MLP:
         model:add(nn.Reshape(64*2*2))
         model:add(nn.Linear(64*2*2, 200))
         model:add(nn.Tanh())
         model:add(nn.Linear(200, noutputs))

      elseif opt.dataset == 'svhn' then
         -- stage 1 : filter bank -> squashing -> L2 pooling -> normalization
         model:add(nn.SpatialConvolutionMM(nfeats, nstates[1], filtsize, filtsize))
         model:add(nn.Tanh())
         model:add(nn.SpatialLPPooling(nstates[1],2,poolsize,poolsize,poolsize,poolsize))
         model:add(nn.SpatialSubtractiveNormalization(nstates[1], normkernel))
   
         -- stage 2 : filter bank -> squashing -> L2 pooling -> normalization
         model:add(nn.SpatialConvolutionMM(nstates[1], nstates[2], filtsize, filtsize))
         model:add(nn.Tanh())
         model:add(nn.SpatialLPPooling(nstates[2],2,poolsize,poolsize,poolsize,poolsize))
         model:add(nn.SpatialSubtractiveNormalization(nstates[2], normkernel))

         -- stage 3 : standard 2-layer neural network
         model:add(nn.Reshape(nstates[2]*filtsize*filtsize))
         model:add(nn.Linear(nstates[2]*filtsize*filtsize, nstates[3]))
         model:add(nn.Tanh())
         model:add(nn.Linear(nstates[3], noutputs))
      end
   end

   -- define loss
   model:add(nn.LogSoftMax())

-- printing the model
print(model)

criterion = nn.ClassNLLCriterion()

-- prepare for training

-- This matrix records the current confusion across classes
confusion = optim.ConfusionMatrix(classes)

-- Retrieve parameters and gradients (1-dim vector)
parameters,gradParameters = model:getParameters()

print(parameters:nElement())

if paths.filep(paramfile) then
   parameters:copy(torch.load(paramfile))
   print('load parameters from file')
end

-- SGD optimizer
optimState = {
   momentum          = opt.momentum,
   weightDecay       = opt.weightDecay,
   learningRate      = opt.learningRate,
   learningRateDecay = opt.learningRateDecay
}
optimMethod = optim.sgd

-- load/initialize epoch
if paths.filep(epochfile) then
   epoch = torch.load(epochfile)
else
   epoch = 1
end
print('epoch: ' .. epoch)
 
function train(node)
   -- epoch tracker
   epoch = epoch or 1

   -- local vars
   local time = sys.clock()

   -- set model to training mode (for modules that differ in training and testing, like Dropout)
   model:training()

   -- shuffle at each epoch
   local shuffle = torch.randperm(trainData.shardsize())

   -- do opt.epochFraction of one epoch
   for t = 1, math.ceil(opt.epochFraction * trainData:shardsize()), opt.batchSize do
      -- create mini batch
      local inputs = {}
      local targets = {}
      for i = t, math.min(t+opt.batchSize-1,trainData:shardsize()) do
         -- load new sample
         local inx = trainData.shard[node][shuffle[i]]
         local input = trainData.data[inx]
         local target = trainData.labels[inx]

         if opt.type == 'double' then input = input:double() end

         table.insert(inputs, input)
         table.insert(targets, target)
      end

      -- create closure to evaluate f(X) and df/dX
      local feval = function(x)
                       -- get new parameters
                       if x ~= parameters then
                          parameters:copy(x)
                       end

                       -- reset gradients
                       gradParameters:zero()

                       -- f is the average of all criterions
                       local f = 0

                       -- evaluate function for complete mini batch
                       for i = 1,#inputs do
                          -- estimate f
                          local output = model:forward(inputs[i])
                          local err = criterion:forward(output, targets[i])
                          f = f + err

                          -- estimate df/dW
                          local df_do = criterion:backward(output, targets[i])
                          model:backward(inputs[i], df_do)

                          -- update confusion
                          confusion:add(output, targets[i])
                       end

                       -- normalize gradients and f(X)
                       gradParameters:div(#inputs)
                       f = f/#inputs

                       -- return f and df/dX
                       return f,gradParameters
                    end

      -- optimize on current mini-batch
      optimMethod(feval, parameters, optimState)
   end

   -- accuracy
   accuracy = confusion.mat:diag():sum() / confusion.mat:sum()

   -- time taken
   time = sys.clock() - time
   traintime = time * 1000 / trainData.shardsize()

   -- next epoch
   confusion:zero()
   epoch = epoch + 1
 
   return accuracy
end

function test()
   -- local vars
   local time = sys.clock()

   -- set model to evaluate mode
   model:evaluate()

   -- test over test data
   for t = 1,testData:size() do
      -- get new sample
      local input = testData.data[t]
      if opt.type == 'double' then input = input:double() end
      local target = testData.labels[t]

      -- test sample
      local pred = model:forward(input)
      confusion:add(pred, target)
   end

   -- timing
   time = sys.clock() - time
   time = time / testData:size()

   accuracy = confusion.mat:diag():sum() / confusion.mat:sum()
   testtime = time * 1000

   -- next iteration:
   confusion:zero()

   return accuracy
end

